/*
 * xwinx.c - X11 client to allow remote access to Win32 desktop
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <windows.h>
#include <wingdi.h>
#include <X11/Xlib.h>
#include <X11/XAtom.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/keysymdef.h>

int pmx, pmy;	/* previous mouse coords */
Display *dpy;
Window win;

/* update the cursor position */
void updCursor (){
	Window rrt, crt;	/* root and client windows cursor is in	*/
	int cmx, cmy;		/* client window mouse coords 		*/
	int rmx, rmy;		/* root window mouse coords 		*/
	unsigned int mrt;	/* mouse return 			*/

	/* Get location of X cursor */
	XQueryPointer(
		dpy,
		win,
		&rrt,
		&crt,
		&rmx,
		&rmy,
		&cmx,
		&cmy,
		&mrt
	);

	/* SetCursorPos() if the curso has moved */
	if ( (pmx != cmx) || (pmy != cmy) ){
		SetCursorPos(cmx, cmy);
		pmx = cmx;
		pmy = cmy;
	}
}

/* convert between an xlib keycode and a win32 virtual keycode */
int convKeycode (int xkey)
{
	KeySym xsym;
	int wkey = 0;
	TCHAR wchr = (TCHAR)0;

	xsym = XKeycodeToKeysym(dpy, xkey, 0);

	/* there has to be a better way... */
	switch (xsym)
	{
		case XK_Alt_L:
			wkey = VK_MENU;
			break;
		case XK_Alt_R:
			wkey = VK_MENU;
			break;
		case XK_Control_L:
			wkey = VK_CONTROL;
			break;
		case XK_Control_R:
			wkey = VK_CONTROL;
			break;
		case XK_Shift_L:
			wkey = VK_SHIFT;
			break;
		case XK_Shift_R:
			wkey = VK_SHIFT;
			break;
		case XK_BackSpace:
			wkey = VK_BACK;
			break;
		case XK_Tab:
			wkey = VK_TAB;
			break;
		case XK_Caps_Lock:
			wkey = VK_CAPITAL;
			break;
		case XK_Return:
			wkey = VK_RETURN;
			break;
		case XK_Escape:
			wkey = VK_ESCAPE;
			break;
		case XK_Delete:
			wkey = VK_DELETE;
			break;
		case XK_Left:
			wkey = VK_LEFT;
			break;
		case XK_Right:
			wkey = VK_RIGHT;
			break;
		case XK_Up:
			wkey = VK_UP;
			break;
		case XK_Down:
			wkey = VK_DOWN;
			break;
		case XK_Page_Up:
			wkey = VK_PRIOR;
			break;
		case XK_Page_Down:
			wkey = VK_NEXT;
			break;
		case XK_Home:
			wkey = VK_HOME;
			break;
		case XK_End:
			wkey = VK_END;
			break;
		case XK_Print:
			wkey = VK_SNAPSHOT;
			break;
		case XK_space:
			wkey = VK_SPACE;
			break;	
		case XK_plus:
			wchr = '+';
			break;
		case XK_comma:
			wchr = ',';
			break;
		case XK_minus:
			wchr = '-';
			break;
		case XK_period:
			wchr = '.';
			break;
		case XK_semicolon:
			wchr = ';';
			break;
		case XK_apostrophe:
			wchr = '\'';
			break;
		case XK_slash:
			wchr = '/';
			break;
		case XK_grave:
			wchr = '`';
			break;
		case XK_bracketleft:
			wchr = '[';
			break;
		case XK_bracketright:
			wchr = ']';
			break;
		case XK_backslash:
			wchr = '\\';
			break;
		case XK_equal:
			wchr = '=';
			break;
		default:
			/* where the key is inside a range that is sequential
			 * in the win32 and X11 keycodes, just change the 
			 * offsett to be relative to the windows offset
			 */
			if (xsym >= XK_A && xsym <= XK_Z)
				wkey = xsym - XK_A + 0x41;
			if (xsym >= XK_a && xsym <= XK_z)
				wkey = xsym - XK_a + 0x41;
			if (xsym >= XK_0 && xsym <= XK_9)
				wkey = xsym - XK_0 + 0x30;
			if (xsym >= XK_KP_0 && xsym <= XK_KP_9)
				wkey = xsym - XK_KP_0 + VK_NUMPAD0;
			if (xsym >= XK_F1 && xsym <= XK_F24)
				wkey = xsym - XK_F1 + VK_F1;
			break;

	}

	if (wchr)
		wkey = VkKeyScan(wchr);

	if (wkey)
	{
		/* be quiet
		printf(
			"Converted %s to a windows scancode.\n",
			XKeysymToString(xsym)
		);
		*/
	} else
	{
		printf(
			"ERROR: XWinX doesn't support the %s key yet.\n",
			XKeysymToString(xsym)
		);
	}

	return(wkey);
}

int WINAPI WinMain(
	HINSTANCE vhin,
	HINSTANCE vpin,
	LPSTR vcmd,
	int vcms
)
{
	HDC dev;
	int xpix;
	int ypix;
	int ret = 0;

	GC con;
	XGCValues vls;
	XImage *img;
	XSetWindowAttributes att;
	long *buf, *fub, *xib;
	int dep;
	BITMAPINFO bmi;
	XEvent evt;
	DWORD flg = 0;		/* flag to use with mouse_event */

	int blocksiz = 10; //50;
	int blocknum = 0;
	int blockcur = 0;
	int blocktop = 0;
	int blockbot = 0;
	int blockpix = 0;
	int blocktot;
	int blockbyt;
 
	/* connect to X11 server */
	dpy = XOpenDisplay(NULL);
	if (!dpy)
	{
		fprintf(
			stderr,
			"Error: Can't open display: %s\n",
			getenv("DISPLAY")
		);
		exit(EXIT_FAILURE);
	}

	/* get depth of X server */
	dep = DefaultDepth(dpy, DefaultScreen(dpy)); 

	/* exit if the remote X server isn't 24 bit */
	if (dep != 24)
	{
		fprintf(stderr, "Error: X server doesn't do 24bpp.\n");
		exit(EXIT_FAILURE);
	}
	
	/* get the size of the windows desktop */
	dev = GetDC(NULL);
	xpix = GetDeviceCaps(dev, HORZRES);
	ypix = GetDeviceCaps(dev, VERTRES);

	printf("XWinX 0.01 Copyright (C) 2001 Kim Saunders\n");
	printf("Local display: %i x %i\n", xpix, ypix);

	att.background_pixel	= BlackPixel(dpy, DefaultScreen(dpy));
	att.border_pixel	= WhitePixel(dpy, DefaultScreen(dpy));
	
	/* create a window on the X server */
	win = XCreateWindow(
		dpy,
		DefaultRootWindow(dpy),
		0,
		0,
		(unsigned int)xpix,
		(unsigned int)ypix,
		0,
		dep,
		InputOutput,
		CopyFromParent,
		CWBackPixel | CWBorderPixel,
		&att
	);

	/* get mouse click and keyboard button events */
	XSelectInput(
		dpy,
		win,
		ExposureMask |
		ButtonPressMask |
		ButtonReleaseMask |
		KeyPressMask |
		KeyReleaseMask
	);

	XMapWindow(dpy, win);
	con = XCreateGC(dpy, win, 0, &vls);
	XFlush(dpy);

	blockbyt = xpix * blocksiz * 4;
	xib = malloc(xpix * ypix * 4);

	/* setup blocks */
	while ( (blocknum * blocksiz) < ypix ){
		blocknum++;
	}
	fub = malloc(blocknum * 4);
	buf = malloc(blocknum * 4);
	for (blockcur = 0; blockcur < blocknum; blockcur++)
	{
		fub[blockcur] = (int)malloc(blockbyt);
		buf[blockcur] = (int)malloc(blockbyt);
	}	

	/* make an image to put the screendump into */
	img = XCreateImage(
		dpy,
		(Visual *)DefaultScreen(dpy),
		dep,
		ZPixmap,
		0,
		(char *)xib,
		xpix,
		ypix,
		32,
		xpix * 4
	);

	XInitImage(img);

	/* Bitmap header describing the bitmap we want to get from GetDIBits */
	bmi.bmiHeader.biSize		= 120;
	bmi.bmiHeader.biWidth		= xpix;
	bmi.bmiHeader.biHeight		= -abs(ypix);
	bmi.bmiHeader.biPlanes		= 1;
	bmi.bmiHeader.biBitCount	= 32;
	bmi.bmiHeader.biCompression	= BI_RGB;
	bmi.bmiHeader.biSizeImage	= 0;
	bmi.bmiHeader.biXPelsPerMeter	= 0;
	bmi.bmiHeader.biYPelsPerMeter	= 0;
	bmi.bmiHeader.biClrUsed		= 0;
	bmi.bmiHeader.biClrImportant	= 0;

	
	while (1)
	{
		int next = -1;
		int sext = -1;

		blocktot = 0;

		/* next and sext are north and south extremes to transfer. next
		 * will be set to the top line number of the area that needs to
		 * be transferred, and sext will be set to the bottom line
		 * number of the area that needs to be transferred.
		 */

		/* if we have been exposed, update the entire screen */
		while ( XCheckMaskEvent(dpy, ExposureMask, &evt) )
		{
			if (evt.type == Expose)
			{
				next = 0;
				sext = ypix - 1;
			}

		}
		
		/* if we haven't been exposed, see what needs updating (because
		 * it has chagned since the last update)
		 */
		if (next == -1)
		{
			/* dump the bitmap from the desktop into our XImage */
			for (blockcur = 0; blockcur < blocknum; blockcur++)
			{
				/* calculate the top and bottom pixels for the
				 * current block 
				 */
				blocktop = blocksiz * blockcur;
				blockbot = (blocksiz * (blockcur + 1)) - 1;
				if (blockbot > ypix - 1)
					blockbot = ypix - 1;

				blockpix = blockbot - blocktop;
				blockpix += 1;
				blocktot += blockpix;

				/* capture the block into buf[blockcur] */
				ret = GetDIBits(
					dev,
					GetCurrentObject(dev, OBJ_BITMAP),
					blocktop,
					blockpix,
					(void *)buf[blockcur],
					&bmi,
					DIB_RGB_COLORS
				);
			
				/* find out if block has changed */
				if (
					memcmp(
						(void *)buf[blockcur],
						(void *)fub[blockcur],
						blockbyt
					) != 0
				)
				{
					memcpy(
						(void *)fub[blockcur],
						(void *)buf[blockcur],
						blockbyt
					);

					if (next == -1)
						next = blocktop;
					sext = blockbot;
				}
				/* sync the windows mouse cursor position with
				 * the X cursor position
				 */
				updCursor();
			}
		}

		if ( next != -1 )
		{
			/* a block from next to sext needs to be transferred
			 * to the X server from the desktop
			 */
			GetDIBits(
				dev,
				GetCurrentObject(dev, OBJ_BITMAP),
				next,
				sext - next + 1,
				xib,
				&bmi,
				DIB_RGB_COLORS
			);
			XPutImage(
				dpy,
				win,
				con,
				img,
				0,
				0,
				0,
				ypix - sext + 1,
				xpix,
				sext - next + 1	
			);
		}
		
		/* process events from the X server */
		while (
			XCheckMaskEvent(
				dpy,
				ButtonPressMask |
				ButtonReleaseMask |
				KeyPressMask |
				KeyReleaseMask,
				&evt
			)
		)
		{
			/* simulate mouse click */
			if (evt.type == ButtonPress)
			{
				if (evt.xbutton.button == Button1)
					flg = MOUSEEVENTF_LEFTDOWN;
				if (evt.xbutton.button == Button2)
					flg = MOUSEEVENTF_MIDDLEDOWN;
				if (evt.xbutton.button == Button3)
					flg = MOUSEEVENTF_RIGHTDOWN;
			}
			if (evt.type == ButtonRelease)
			{
				if (evt.xbutton.button == Button1)
					flg = MOUSEEVENTF_LEFTUP;
				if (evt.xbutton.button == Button2)
					flg = MOUSEEVENTF_MIDDLEUP;
				if (evt.xbutton.button == Button3)
					flg = MOUSEEVENTF_RIGHTUP;
			}				
			
			if (
				evt.type == ButtonPress ||
				evt.type == ButtonRelease
			)
			{
				SetCursorPos(evt.xbutton.x, evt.xbutton.y);
				mouse_event(flg, 0, 0, 0, (int)NULL);
			}
			
			/* simulate a key press */
			if (evt.type == KeyPress || evt.type == KeyRelease)
			{
				int vkc;		/* virtual keycode */

				vkc = convKeycode(evt.xkey.keycode);
				
				keybd_event(
					vkc,
					MapVirtualKey(vkc, 0),
					evt.type == KeyRelease ?
					KEYEVENTF_KEYUP : (DWORD)NULL,
					0
				);
			}
		}
	}

	XCloseDisplay(dpy);
	return(EXIT_SUCCESS);
}

